from dataclasses import dataclass

from textual.message import Message
from textual.widgets import ListItem


@dataclass
class ContainersAndImagesExpaned(Message, bubble=True):
    pass


@dataclass
class ClickedContainer(Message, bubble=True):
    clicked_container: ListItem



import re
import time
from threading import Event, Thread
from typing import Any

from textual import on
from textual.app import ComposeResult
from textual.containers import Horizontal
from textual.widget import Widget
from textual.widgets import (
    Input,
    Static,
    Switch,
    TabbedContent,
    TabPane,
)

from application.docker_manager import DockerManager
from application.widget.log_viewer import LogLines
from application.widget.shell import ShellPane
from application.widget.statistics import Statistics


class ContentWindow(Widget):
    current_list_index = None
    search_keyword = ""
    matches, indices = None, None
    log_task_stop_event = Event()

    def __init__(
        self,
        docker_manager: DockerManager,
        *children: Widget,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        self.docker_manager = docker_manager
        super().__init__(
            *children, name=name, id=id, classes=classes, disabled=disabled
        )

    def compose(self) -> ComposeResult:
        global logs

        yield Horizontal(
            Input(id="search_log_input", placeholder="Search logs...", type="text"),
            Horizontal(
                Static("case-sensitive", classes="label"),
                Switch(id="case-sensitive-switch", animate=True),
                classes="case-sensitive-switch",
            ),
            classes="container",
        )
        with TabbedContent():
            with TabPane("Logs", id="logpane"):
                logs = LogLines(id="logs", highlight=True, auto_scroll=True, name="Log")
                logs.border_title = self.docker_manager.selected_container.name
                logs.scroll_end(animate=False)
                yield logs
            with TabPane("Attributes", id="attributespane"):
                attributes = LogLines(
                    id="attributes_log",
                    highlight=True,
                    auto_scroll=True,
                    name="Attributes",
                )
                attributes.border_title = self.docker_manager.selected_container.name
                attributes.scroll_end(animate=False)
                yield attributes
            with TabPane("Environment", id="environmentpane"):
                environment = LogLines(
                    id="environment_log",
                    highlight=True,
                    auto_scroll=True,
                    name="Environment",
                )
                environment.border_title = self.docker_manager.selected_container.name
                environment.scroll_end(animate=False)
                yield environment
            with TabPane("Statistics", id="statisticspane"):
                yield Statistics(label="CPU (%)", id="statistics_plot_cpu")
                yield Statistics(label="RAM (MB)", id="statistics_plot_memory")

                statistics = LogLines(
                    id="statistics_log",
                    highlight=True,
                    auto_scroll=True,
                    name="Statistics",
                )
                statistics.border_title = self.docker_manager.selected_container.name
                statistics.scroll_end(animate=False)
                yield statistics
            yield ShellPane(
                title="Shell", id="shellpane", docker_manager=self.docker_manager
            )

    def search_logs(self, pattern):
        case_sensitive_switch = self.query_one("#case-sensitive-switch", Switch).value
        case_sensitive = 0 if case_sensitive_switch else re.IGNORECASE
        compiled_pattern = re.compile(pattern, case_sensitive)

        current_tab = self.query_one(TabbedContent).active
        self.current_content_window: LogLines = self.query_one(
            f"#{current_tab}"
        ).children[0]
        results = [
            (line, i)
            for i, line in enumerate(self.current_content_window.lines)
            if compiled_pattern.search(line.text)
        ]

        if results:
            self.matches, self.indices = zip(*results)
            self.matches = list(self.matches)
            self.indices = list(self.indices)
        else:
            self.matches = []
            self.indices = []

    @on(Input.Changed)
    def input_changed(self, input=Input()) -> None:
        if self._shell_tab_active():
            return  # No text to look for when inside shell

        keyword = input.value
        self._update_case_sensitivity()

        if not keyword or keyword != self.search_keyword:
            self._reset_search_state()

        if self.current_list_index is None and keyword:
            self._perform_search(keyword)

        if self.indices:
            self._update_input_display(keyword)
        else:
            self._show_no_results()

    @on(Input.Submitted)
    def input_submitted(self, input=Input()) -> None:
        if self._shell_tab_active():
            return  # No text to look for when inside shell

        keyword = input.value
        self._update_case_sensitivity()
        self._update_list_index()
        self._update_input_display(keyword)

    def _update_case_sensitivity(self) -> None:
        case_sensitive_switch = self.query_one("#case-sensitive-switch", Switch).value
        logs.case_sensitive = 0 if case_sensitive_switch else re.IGNORECASE

    def _reset_search_state(self) -> None:
        self.current_list_index = None
        self.search_keyword = ""
        self.indices = []

    def _perform_search(self, keyword: str) -> None:
        self.search_logs(keyword)
        self.search_keyword = keyword
        self.current_list_index = len(self.indices) - 1

    def _update_input_display(self, keyword: str) -> None:
        try:
            self.update_input_border()
        except IndexError:
            return  # No matches, don't update

        self.current_content_window.scroll_to(
            y=self._get_indices_list_value(), animate=False, duration=0
        )
        self.current_content_window.keyword = keyword
        self.current_content_window.current_index = self._get_indices_list_value()
        self.current_content_window.refresh()  # Force refresh, just to get rid of slow visual changes.

    def _show_no_results(self) -> None:
        input = self.query_one("#search_log_input")
        input.border_title = "No results"
        self.current_list_index = 0

    def _update_list_index(self) -> None:
        if self.current_list_index == 0:
            self.current_list_index = len(self.indices) - 1
        else:
            self.current_list_index -= 1

    def _get_indices_list_value(self):
        return self.indices[self.current_list_index]

    def update_input_border(self) -> None:
        line = str(self._get_indices_list_value())
        input = self.query_one("#search_log_input")
        input.border_title = f"{self.current_list_index + 1}/{len(self.indices)}"
        input.border_subtitle = f"line {line}"

    def live_logs_task(self):
        self.docker_manager.live_container_logs(logs, self.log_task_stop_event)

    def run_log_task(self):
        self.log_task_stop_event.set()  # Signal any existing task to stop
        self.log_task_stop_event = Event()  # Create a new stop event for the new task
        Thread(target=self.live_logs_task, daemon=True).start()

    def live_statistics_task(self):
        stop_event = Event()

        while not stop_event.is_set():
            try:
                stats = self._fetch_container_stats()
                cpu, memory = self._parse_stats(stats)
                self._update_plots_if_visible(cpu, memory)
                self._update_logs(cpu, memory)
            except Exception:
                stop_event.set()
                self.live_logs_task()

    def _fetch_container_stats(self):
        return self.docker_manager.selected_container.stats(stream=False)

    def _parse_stats(self, stats: dict[Any, Any]):
        try:
            cpu_usage = (
                stats["cpu_stats"]["cpu_usage"]["total_usage"]
                / stats["cpu_stats"]["system_cpu_usage"]
            )
            memory_usage_mb = stats["memory_stats"]["usage"] / (1024 * 1024)
            cpu = f"{cpu_usage * 100:.3f}%"
            memory = f"{memory_usage_mb:.3f} MB"
            return cpu, memory
        except KeyError:
            # TODO: Correctly handle this case.
            return "N/A", "N/A"

    def _update_plots_if_visible(self, cpu: str, memory: str):
        if self.query_one(TabbedContent).active == "statisticspane":
            cpu = float(cpu.replace("%", ""))
            memory = float(memory.replace("MB", ""))
            current_time = time.strftime("%M:%S")
            self.query_one("#statistics_plot_cpu", Statistics).update(cpu, current_time)
            self.query_one("#statistics_plot_memory", Statistics).update(
                memory, current_time
            )

    def _update_logs(self, cpu: str, memory: str):
        logs.border_subtitle = f"cpu: {cpu} | ram: {memory} | logs: {len(logs.lines)}"

    def _shell_tab_active(self):
        return self.query_one(TabbedContent).active == "shellpane"



import datetime
import os
import re
import subprocess
from pathlib import Path

import requests
import yaml
from colorama import Fore, Style
from packaging.version import parse as version_parse
from pydantic import BaseModel

POCKER_CONFIG_BASE_PATH = Path.home() / ".config/pocker"


def get_latest_version():
    response = requests.get(
        "https://api.github.com/repos/pommee/Pocker/tags?per_page=1"
    )

    if response.status_code == 200:
        response: dict = response.json()[0]
        latest_version = response.get("name")
        return version_parse(latest_version)
    else:
        return None


def get_current_version():
    try:
        result = subprocess.run(
            ["pipx", "list", "--short"], text=True, capture_output=True, check=True
        )

        for line in result.stdout.splitlines():
            if "pocker-tui" in line:
                version = line.split()[1]
                return version

    except subprocess.CalledProcessError:
        return None


class Version_Fetch(BaseModel):
    time_fetched: datetime.datetime
    version_fetched: str


def write_latest_version_fetch(version):
    version_fetch_dict = {
        "time_fetched": datetime.datetime.now(),
        "version_fetched": version,
    }
    if not os.path.exists(POCKER_CONFIG_BASE_PATH):
        POCKER_CONFIG_BASE_PATH.mkdir()

    with open(POCKER_CONFIG_BASE_PATH / "latest_version_fetch.yaml", "w") as file:
        yaml.safe_dump(version_fetch_dict, file)


def read_latest_version_fetch():
    if not os.path.exists(POCKER_CONFIG_BASE_PATH / "latest_version_fetch.yaml"):
        latest_version = get_latest_version()
        write_latest_version_fetch(latest_version.base_version)
    with open(POCKER_CONFIG_BASE_PATH / "latest_version_fetch.yaml", "r") as file:
        fetch_dict = yaml.safe_load(file)
        return Version_Fetch(**fetch_dict)


def time_since_last_fetch():
    return (
        datetime.datetime.now() - read_latest_version_fetch().time_fetched
    ).total_seconds() / 60


commit_and_link_pattern = (
    r"\[([a-f0-9]+)\]\((https://github\.com/pommee/Pocker/commit/[a-f0-9]+)\)"
)
changed_line = r"\(\[([a-f0-9]+)\]\(([^)]+)\)\)"


def bright_text(text):
    return f"{Style.BRIGHT}{text}{Style.RESET_ALL}"


class Release:
    def __init__(self, unparsed_text: dict) -> None:
        self.unparsed_text = unparsed_text
        self.version = None
        self.features = None
        self.bug_fixes = None

    def extract_version(self):
        version_pattern = r"# \[(\d+\.\d+\.\d+)\]"
        version_match = re.search(version_pattern, self.unparsed_text)
        self.version = version_match.group(1) if version_match else None

    def extract_features(self):
        features_pattern = r"### Features\n\n((?:\* .+\n?)+)"
        features_match = re.search(features_pattern, self.unparsed_text)
        self.features = (
            features_match.group(1).strip().split("\n") if features_match else []
        )

    def extract_bug_fixes(self):
        bug_fixes_pattern = r"### Bug Fixes\n\n((?:\* .+\n?)+)"
        bug_fixes_match = re.search(bug_fixes_pattern, self.unparsed_text)
        self.bug_fixes = (
            bug_fixes_match.group(1).strip().split("\n") if bug_fixes_match else []
        )

    def extract_headers(self):
        self.extract_version()
        self.extract_features()
        self.extract_bug_fixes()

    def print_changelog_header(self):
        print(bright_text(f"\nv{self.version} Changelog\n"))

    def print_features(self):
        if len(self.features) == 0:
            return
        print(bright_text("Features:\n"))
        for feature in self.features:
            processed_text = re.sub(changed_line, r"[\1](\2)", feature).replace("*", "")
            match = re.search(commit_and_link_pattern, processed_text)
            if match:
                commit_hash = f"[{match.group(1)}]"
                url = match.group(2)
                print(
                    f" - {Fore.YELLOW}{commit_hash}{Style.RESET_ALL} {processed_text.replace(commit_hash, '').replace(f'({url})', '')}"
                )

    def print_bug_fixes(self):
        if len(self.bug_fixes) == 0:
            return
        print(bright_text("\nBug Fixes:\n"))
        for bug_fix in self.bug_fixes:
            processed_text = re.sub(changed_line, r"[\1](\2)", bug_fix).replace("*", "")
            match = re.search(commit_and_link_pattern, processed_text)
            if match:
                commit_hash = f"[{match.group(1)}]"
                url = match.group(2)
                print(
                    f" - {Fore.YELLOW}{commit_hash}{Style.RESET_ALL} {processed_text.replace(commit_hash, '').replace(f'({url})', '')}"
                )


def read_changelog(previously_installed_version):
    response = requests.get("https://api.github.com/repos/pommee/Pocker/releases")
    changelog: dict = response.json()
    releases: list[Release] = []

    for change in changelog:
        change_text = change.get("body")
        release = Release(change_text)

        release.extract_headers()
        releases.append(release)

    for release in releases:
        current_release_version = version_parse(release.version)

        if current_release_version > previously_installed_version:
            release.print_changelog_header()
            release.print_features()
            release.print_bug_fixes()

    print("")


def update_changelog():
    response = requests.get(
        "https://raw.githubusercontent.com/pommee/Pocker/main/CHANGELOG.md"
    )
    if response.status_code == 200:
        file_path = POCKER_CONFIG_BASE_PATH / "CHANGELOG.md"
        with open(file_path, "w") as file:
            file.write(response.text)



from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Footer, Markdown


class StartupError(ModalScreen):
    BINDINGS = [
        Binding(key="q", action="quit", description="Quit"),
    ]

    def __init__(
        self,
        error_reason: Exception,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self.error_reason = error_reason
        super().__init__(name, id, classes)

    def compose(self) -> ComposeResult:
        ERROR_HEADER = "# ERROR"
        REASON_HEADER = "### Reason"

        with VerticalScroll():
            yield Markdown(ERROR_HEADER)
            yield Markdown(f"{REASON_HEADER}: {self.error_reason.REASON}")

            HELP_MARKDOWN = Markdown(id="help")
            HELP_MARKDOWN._markdown = self.error_reason.HELP
            HELP_MARKDOWN.border_title = "HELP"
            yield HELP_MARKDOWN

        yield Footer()

    async def action_quit(self):
        await self.app.action_quit()



import logging
from threading import Event
import threading

import docker
from docker.models.containers import Container
from docker.models.images import Image, ImageCollection
from textual.logging import TextualHandler

from application.util.config import CONFIG_PATH, Config
from application.widget.log_viewer import LogLines

logging.basicConfig(
    level="INFO",
    handlers=[TextualHandler()],
)


class NoVisibleContainers(Exception):
    REASON = "No container(s) to display"
    HELP = f"""
- No containers are running.
- Set `show_all_containers` to true in config to display exited containers.  

Config can be found at `{CONFIG_PATH}`
"""
    pass


class DockerManager:
    def __init__(self, config: Config) -> None:
        self.client = docker.from_env()
        self.config = config
        self.containers: dict[str, Container] = {}
        self.images: ImageCollection = None
        self.selected_container: Container = None
        self.selected_image = None

        container_thread = threading.Thread(target=self._load_containers)
        image_thread = threading.Thread(target=self._load_images)

        container_thread.start()
        image_thread.start()

        container_thread.join()

        if not self.containers:
            raise NoVisibleContainers()

        self.selected_container = next(iter(self.containers.values()))

    def _load_containers(self):
        containers = self.client.containers.list(all=self.config.show_all_containers)
        self.containers = {container.name: container for container in containers}

    def _load_images(self):
        self.images = self.client.images.list(all=True)

    @property
    def attributes(self) -> Container:
        return self.selected_container.attrs

    @property
    def environment(self) -> dict:
        return self.selected_container.attrs.get("Config").get("Env")

    @property
    def statistics(self) -> Image:
        return self.selected_container.stats(stream=False)

    def append_container(self, container_name: str):
        self.containers[container_name] = self.client.containers.get(container_name)

    def logs(self):
        logs: bytes = self.selected_container.logs(
            tail=self.config.log_tail, follow=False, stream=False
        )
        return logs.decode("utf-8").strip()

    def status(self, container: Container):
        status = "[U]"
        if container.status == "running":
            status = "running"
        else:
            status = "down"

        return status

    def live_container_logs(self, logs: LogLines, stop_event: Event):
        logs.clear()

        try:
            log_stream = self.selected_container.logs(
                stream=True, follow=True, tail=self.config.log_tail
            )

            for log in log_stream:
                if stop_event.is_set():
                    break
                logs.write(log.decode("utf-8").rstrip())

        except Exception:
            stop_event.set()  # Handle exceptions, for example, if the container is removed

        finally:
            try:
                log_stream.close()
            except ValueError:
                pass



from __future__ import annotations

from textual.reactive import var
from textual_plotext import PlotextPlot


class Statistics(PlotextPlot):
    marker: var[str] = var("braille")

    def __init__(
        self,
        label: str,
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.label = label
        self._data: list[float] = []
        self._time: list[str] = []

    def on_mount(self) -> None:
        """Initialize the plot."""
        self.plt.date_form("M:S")

    def replot(self) -> None:
        """Redraw the plot with updated usage data."""
        self.plt.clear_data()
        self.plt.plot(self._time, self._data, marker=self.marker, label=self.label)
        self.refresh()

    def update(self, data: str, timestamp: str) -> None:
        """Update the data for the usage plot.

        Args:
            data: The current value for a specific measurement.
            timestamp: The current time as a string.
        """
        self._data.append(float(data))
        self._time.append(timestamp)
        self.replot()

    def _watch_marker(self) -> None:
        """React to the marker being changed."""
        self.replot()



import webbrowser

from textual import on
from textual.app import ComposeResult
from textual.containers import Grid, Horizontal, Vertical, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Markdown, Static, Switch

from application.util.config import load_config

HELP_MD = """
\n# Settings \n
Consult the [README](https://github.com/pommee/Pocker/blob/main/README.md) to get more information about certain settings.

Press **ESC** to close this window.

---

"""


class SettingsScreen(ModalScreen):
    BINDINGS = [
        ("escape", "dismiss"),
    ]

    def __init__(
        self, name: str | None = None, id: str | None = None, classes: str | None = None
    ) -> None:
        super().__init__(name, id, classes)

    def compose(self) -> ComposeResult:
        config = load_config()

        with VerticalScroll():
            yield Markdown(HELP_MD, id="help", classes="help")

            with Grid():
                with Horizontal(id="container-show-all-containers"):
                    with Vertical():
                        yield Static("Show all containers.", classes="label")
                        with Horizontal(classes="switch-with-status"):
                            yield Switch(
                                id="show-all-containers",
                                animate=True,
                                value=config.show_all_containers,
                            )
                            yield Static(
                                str(config.show_all_containers), classes="switch-status"
                            )

                with Horizontal(id="container-start-fullscreen"):
                    with Vertical():
                        yield Static(
                            "Start in fullscreen. Containers & images won't be visible.",
                            classes="label",
                        )
                        with Horizontal(classes="switch-with-status"):
                            yield Switch(
                                id="start-fullscreen",
                                animate=True,
                                value=config.start_fullscreen,
                            )
                            yield Static(
                                str(config.start_fullscreen), classes="switch-status"
                            )

                with Horizontal(id="container-start-scroll"):
                    with Vertical():
                        yield Static("Start with scrolling.", classes="label")
                        with Horizontal(classes="switch-with-status"):
                            yield Switch(
                                id="start-scroll",
                                animate=True,
                                value=config.start_scroll,
                            )
                            yield Static(
                                str(config.start_scroll), classes="switch-status"
                            )

                with Horizontal(id="container-start-wrap"):
                    with Vertical():
                        yield Static("Start with wrapped logs.", classes="label")
                        with Horizontal(classes="switch-with-status"):
                            yield Switch(
                                id="start-wrap", animate=True, value=config.start_wrap
                            )
                            yield Static(
                                str(config.start_wrap), classes="switch-status"
                            )

    def on_switch_changed(self, value: Switch.Changed):
        config = load_config()
        switch_id = value.switch.id
        new_value = value.value

        field_name = switch_id.replace("-", "_")
        setattr(config, field_name, new_value)
        config.save()

        container = self.query_one(f"#container-{switch_id}")

        if container:
            static_widget: Static = container.query(".switch-status").first()
            static_widget.update(str(new_value))

    @on(Markdown.LinkClicked)
    def on_markdown_link_clicked(self, event: Markdown.LinkClicked) -> None:
        self.action_go(event.href)

    def action_go(self, href: str) -> None:
        webbrowser.open(href)
